/*
MIDI Processing for a Continuum 
 
*/

autowatch = 1;
inlets = 1;
outlets = 1;

var MidiClass = {
    UNKNOWN : 0,
    SYSTEM_REALTIME : 1,
    SYSTEM_COMMON : 2,
    VOICE : 3
};

function MidiMsg( msg ) {
    this.msg = msg;
}

MidiMsg.prototype.Class = function() {
    if( this.msg[0] >= 0xF8 ) {
        return MidiClass.SYSTEM_REALTIME;
    } else if( this.msg[0] >= 0xF0 ) {
        return MidiClass.SYSTEM_COMMON;
    } else if( this.msg[0] >= 0x80 ) {
        return MidiClass.VOICE;
    } else {
        return MidiClass.UNKNOWN;
    }
}

MidiMsg.prototype.Channel = function() {
    if( this.Class() == MidiClass.VOICE ) {
        return this.msg[0] & 0xf;
    }
    return -1;
}

MidiMsg.prototype.IsController = function() {
	return this.msg[0] & 0xF0 == 0xB0;
}

MidiMsg.prototype.Controller = function() {
	if( this.IsController() ) {
		return this.msg[1];
	}
	return -1;
}

MidiMsg.prototype.ControllerValue = function() {
	if( this.msg[0] & 0xF0 == 0xB0 ) {
		return this.msg[2];
	}
	return -1;
}

MidiMsg.prototype.IsPitchWheel = function() {
	return this.msg[0] & 0xF0 == 0xE0;
}

MidiMsg.prototype.PitchWheel = function() {
	if( this.IsPitchWheel() ) {
		return ( this.msg[1] | ( this.msg[2] << 7 ) ) - 8192;
	}
	return -1;
}

MidiMsg.prototype.IsNoteMsg = function() {
	return this.msg[0] & 0xE0 == 0x80;
}

MidiMsg.prototype.NoteOnKeyVel = function() {
	if( ( this.msg[0] & 0xF0 == 0x90 ) && this.msg[2] != 0 ) {
		return this.msg[1,2];
	}
	return -1;
}

MidiMsg.prototype.NoteOffKeyVel = function() {
	if( this.msg[0] & 0xF0 == 0x80 ) {
		return this.msg[1,2];
	} else if( ( this.msg[0] & 0xF0 == 0x90 ) && this.msg[2] == 0 ) {
		return [ this.msg[1], 127 ];
	}
	return -1;
}

function MidiAccumulator() {
    this._offset = 0;
    this._length = 0;
    this._msg = [];
}

MidiAccumulator.prototype._MidiSysCommonBytes = [ -1, 2, 3, 2, -1, -1, 1, -2 ];

MidiAccumulator.prototype.HandleMessage = function( msg ) {
    alert( "This should be replaced with user code" );
}

MidiAccumulator.prototype.HandleError = function( msg ) { }

// Handle the high level control logic for accumulating a MIDI message
MidiAccumulator.prototype.Accumulate = function( val ) {

    // Realtime Messages
    if( val >= 0xF8 ) {
        this._NewStatus( val, 0 );

    // System Common Messages
    } else if( val >= 0xF0 ) {
        this._NewStatus( val, this._MidiSysCommonBytes[ v & 0xf ] );

    // 2-byte Voice Commands
    } else if( val >= 0xC0 && val <= 0xDF ) {
        this._NewStatus( val, 2 );

    // 3-byte Voice Commands
    } else if( val >= 0x80 ) {
        this._NewStatus( val, 3 );

    // Any other message just needs to be accumulated
    } else {
        this._AccumMultiByte( val );
    }
}

// Accumulate a byte in a multi-byte message
MidiAccumulator.prototype._AccumMultiByte = function( val ) {

    // Length -1 = In SysEx and accumulate forever 
    if( this._length == -1 ) {
        this._msg.push( val );

    // Length 2+ = Accumulate another message with the current status
    } else if( this._length > 1 ) {

        // Inject the byte into the multi-byte buffer
        this._msg[this._offset] = val;
        this._offset++;

        // If we hit the correct numer of bytes, handle the message
        if( this._offset == this._length ) {
            this.HandleMessage( new MidiMsg( this._msg ) );
            this._offset = 1;
        }

    // Any other length is an error
    } else {
        this.HandleError( [val] );
    }
}

// Handle the boundaries relating to a new status message
MidiAccumulator.prototype._NewStatus = function ( val, len ) {

    // Length 0 = System Realtime (doesn't effect the current status)
    if( len == 0 ) {
        this.HandleMessage( new MidiMsg( [val] ) );

    // Length -2 = End of a SysEx
    } else if( len == -2 ) {
        this._AccumMultiByte( val );
        this.HandleMessage( this._msg );
        this._msg.length = 1;

    // Handle real status changes here
    } else {
        this._msg = [val];
        this._offset = 1;
        this._length = len;

        // Length 1 = Handle immediately (no data required)
        if( len == 1 ) {
            this.HandleMessage( new MidiMsg( this._msg ) );
            this._length = 1;

        // Any other length means we need to accumulate messages
        } else if( len > 1 ) {
            this._msg.length = len;
        }
    }
}

var accum = new MidiAccumulator();

accum.HandleMessage = function( msg ) {
	outlet( 0, Array("instance", msg.Channel(), "midievent").concat(msg.msg) );
}

accum.HandleError = function( msg ) {
	post( "Error" );
	post( msg );
	post();
}

function anything() {
	val = arrayfromargs(messagename,arguments);
	post("anything was called: " + val.toSource());
	post();
}

function msg_int(v) {
    accum.Accumulate( v );
}

